{------------------------------------------------------------------------------}
{                                                                              }
{  Office Assistant                                                            }
{  by Kambiz R. Khojasteh                                                      }
{                                                                              }
{  kambiz@delphiarea.com                                                       }
{  http://www.delphiarea.com                                                   }
{                                                                              }
{------------------------------------------------------------------------------}

{$I DELPHIAREA.INC}

unit oaDirect;

interface

uses
  Windows, Messages, Classes, Graphics, oaEngine, oaActor, oaTypes;

type

  TRequestQueue = class;

  TAssistantDirector = class(TThread)
  private
    Requests: TRequestQueue;
    RunningID: TRequestID;
    IdleID: TRequestID;
    IdleTimeout: DWORD;
    IdleRetryCount: Byte;
    CS: TRTLCriticalSection;
    ActiveEvent: THandle;
    IdleEvent: THandle;
    fActor: TAssistantActor;
    fEngine: TAssistantEngine;
    fEnabled: Boolean;
    fAutoIdle: Boolean;
    fIdleLevel: TIdleLevel;
    fMinIdlePeriod: Word;
    fMaxIdlePeriod: Word;
    fOnPlayStart: TPlayEvent;
    fOnPlayStop: TPlayEvent;
    fOnIdle: TIdleEvent;
    procedure SetEnabled(Value: Boolean);
    procedure SetAutoIdle(Value: Boolean);
    procedure SetIdleLevel(Value: TIdleLevel);
    procedure ActorLoaded(Sender: TObject);
    procedure ActorUnloaded(Sender: TObject);
    procedure SequenceStarted(Sender: TObject);
    procedure SequenceStopped(Sender: TObject);
    procedure AdjustIdleTimeout;
    procedure PlayRequest;
    procedure PlayIdle;
  protected
    procedure Execute; override;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Terminate;
    function PlayIndex(Index: Word): TRequestID;
    function PlayName(const Name: String): TRequestID;
    function PlayAction(Action: TActorAction): TRequestID;
    function PlayRandom(const ActionList: array of TActorAction): TRequestID;
    procedure Stop(RequestID: TRequestID);
    procedure StopAll;
    function Playing: Boolean;
    function PlayingID: TRequestID;
    function PlayIdling: Boolean;
    function WaitForRequest(RequestID: TRequestID): Boolean;
    function IsValidRequest(RequestID: TRequestID): Boolean;
    function IsPlaying(Actions: TActorActions): Boolean;
    function ActionOfOrientation(Kind: TOrientedAction;
      const Origin, Target: TPoint): TActorAction;
    property Actor: TAssistantActor read fActor;
    property Engine: TAssistantEngine read fEngine;
    property AutoIdle: Boolean read fAutoIdle write SetAutoIdle;
    property IdleLevel: TIdleLevel read fIdleLevel write SetIdleLevel;
    property MinIdlePeriod: Word read fMinIdlePeriod write fMinIdlePeriod;
    property MaxIdlePeriod: Word read fMaxIdlePeriod write fMaxIdlePeriod;
    property Enabled: Boolean read fEnabled write SetEnabled;
    property OnPlayStart: TPlayEvent read fOnPlayStart write fOnPlayStart;
    property OnPlayStop: TPlayEvent read fOnPlayStop write fOnPlayStop;
    property OnIdle: TIdleEvent read fOnIdle write fOnIdle;
  end;

  TRequestQueue = class(TList)
  private
    fEnabled: Boolean;
    LastID: Word;
    NoEmptyEvent: THandle;
    CS: TRTLCriticalSection;
    procedure SetEnabled(Value: Boolean);
    function GetRequests(Index: Integer): TRequestID;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Lock;
    procedure Unlock;
    procedure Clear; {$IFDEF DELPHI4_UP} override; {$ENDIF}
    function Get: TRequestID;
    function Put(AnimationIndex: Word): TRequestID;
    procedure Cancel(RequestID: TRequestID);
    function Query(RequestID: TRequestID): Boolean;
    function WaitFor(Timeout: DWORD): Boolean;
    property Enabled: Boolean read fEnabled write SetEnabled;
    property Requests[Index: Integer]: TRequestID read GetRequests; default;
  end;

procedure WaitForEventAndPaint(Event: THandle);

implementation

procedure WaitForEventAndPaint(Event: THandle);
var
  Msg: TMsg;
begin
  if GetCurrentThreadID = MainThreadID then
    while WaitForSingleObject(Event, 0) = WAIT_TIMEOUT do
    begin
      if PeekMessage(Msg, 0, WM_PAINT, WM_PAINT, PM_REMOVE) then
        DispatchMessage(Msg);
    end
  else
    WaitForSingleObject(Event, INFINITE);
end;

{ TRequestQueue }

constructor TRequestQueue.Create;
begin
  inherited Create;
  InitializeCriticalSection(CS);
  NoEmptyEvent := CreateEvent(nil, True, False, nil);
  fEnabled := True;
end;

destructor TRequestQueue.Destroy;
begin
  inherited Destroy;
  CloseHandle(NoEmptyEvent);
  DeleteCriticalSection(CS);
end;

procedure TRequestQueue.Lock;
begin
  EnterCriticalSection(CS);
end;

procedure TRequestQueue.Unlock;
begin
  LeaveCriticalSection(CS);
end;

procedure TRequestQueue.Clear;
begin
  EnterCriticalSection(CS);
  try
    ResetEvent(NoEmptyEvent);
    inherited Clear;
  finally
    LeaveCriticalSection(CS);
  end;
end;

function TRequestQueue.Get: TRequestID;
begin
  EnterCriticalSection(CS);
  try
    if Count > 0 then
    begin
      Result := Requests[0];
      Delete(0);
      if Count = 0 then ResetEvent(NoEmptyEvent);
    end
    else
      Result := 0;
  finally
    LeaveCriticalSection(CS);
  end;
end;

function TRequestQueue.Put(AnimationIndex: Word): TRequestID;
begin
  Result := 0;
  if Enabled and (AnimationIndex > 0) then
  begin
    EnterCriticalSection(CS);
    try
      Inc(LastID);
      Result := (DWORD(LastID) shl 16) or AnimationIndex;
      if Add(Pointer(Result)) >= 0 then
        SetEvent(NoEmptyEvent)
      else
        Result := 0;
    finally
      LeaveCriticalSection(CS);
    end;
  end;
end;

procedure TRequestQueue.Cancel(RequestID: TRequestID);
var
  Index: Integer;
begin
  EnterCriticalSection(CS);
  try
    Index := IndexOf(Pointer(RequestID));
    if Index >= 0 then Delete(Index);
    if Count = 0 then ResetEvent(NoEmptyEvent);
  finally
    LeaveCriticalSection(CS);
  end;
end;

function TRequestQueue.Query(RequestID: TRequestID): Boolean;
begin
  EnterCriticalSection(CS);
  try
    Result := (IndexOf(Pointer(RequestID)) >= 0);
  finally
    LeaveCriticalSection(CS);
  end;
end;

function TRequestQueue.WaitFor(Timeout: DWORD): Boolean;
begin
  Result := (WaitForSingleObject(NoEmptyEvent, Timeout) <> WAIT_TIMEOUT);
end;

procedure TRequestQueue.SetEnabled(Value: Boolean);
begin
  if Enabled <> Value then
  begin
    EnterCriticalSection(CS);
    try
      fEnabled := Value;
      if Enabled then
        ResetEvent(NoEmptyEvent)
      else
      begin
        inherited Clear;
        SetEvent(NoEmptyEvent);
      end;
    finally
      LeaveCriticalSection(CS);
    end;
  end;
end;

function TRequestQueue.GetRequests(Index: Integer): TRequestID;
begin
  EnterCriticalSection(CS);
  try
    if (Index >= 0) and (Index < Count) then
      Result := TRequestID(inherited Get(Index))
    else
      Result := 0;
  finally
    LeaveCriticalSection(CS);
  end;
end;

{ TAssistantDirector }

constructor TAssistantDirector.Create;
begin
  inherited Create(True);
  InitializeCriticalSection(CS);
  ActiveEvent := CreateEvent(nil, True, False, nil);
  IdleEvent := CreateEvent(nil, True, True, nil);
  Requests := TRequestQueue.Create;
  Requests.Enabled := False;
  fActor := TAssistantActor.Create;
  fActor.OnLoad := ActorLoaded;
  fActor.OnUnload := ActorUnloaded;
  fEngine := TAssistantEngine.Create(Actor);
  fEngine.OnSequenceStart := SequenceStarted;
  fEngine.OnSequenceStop := SequenceStopped;
  fAutoIdle := True;
  fMinIdlePeriod := 1000;
  fMaxIdlePeriod := 8000;
  IdleLevel := Low(TIdleLevel);
  Resume;
end;

destructor TAssistantDirector.Destroy;
begin
  fActor.Free;
  fEngine.Free;
  Requests.Free;
  CloseHandle(IdleEvent);
  CloseHandle(ActiveEvent);
  DeleteCriticalSection(CS);
  inherited Destroy;
end;

procedure TAssistantDirector.Terminate;
begin
  Engine.Terminate;
  Engine.WaitFor;
  inherited Terminate;
  EnterCriticalSection(CS);
  try
    SetEvent(IdleEvent);
    Enabled := False;
    SetEvent(ActiveEvent);
  finally
    LeaveCriticalSection(CS);
  end;
end;

procedure TAssistantDirector.Execute;
var
  AnyRequest: Boolean;
begin
  while not Terminated do
  begin
    WaitForSingleObject(ActiveEvent, INFINITE);
    if Enabled then
    begin
      AnyRequest := Requests.WaitFor(IdleTimeout);
      if Requests.Enabled then
      begin
        if AnyRequest then
        begin
          if (RunningID <> 0) and (RunningID = IdleID) then
            Engine.Sequencing := False;
          WaitForSingleObject(IdleEvent, INFINITE);
          if Requests.Enabled then
            PlayRequest;
          if (RunningID = 0) or (RunningID <> IdleID) then
            WaitForSingleObject(IdleEvent, INFINITE);
        end
        else if WaitForSingleObject(IdleEvent, 0) <> WAIT_TIMEOUT then
          PlayIdle;
      end;
    end;
  end;
end;

procedure TAssistantDirector.PlayRequest;
begin
  EnterCriticalSection(CS);
  try
    if (RunningID <> 0) and Engine.PrepareToReturn then
    begin
      RunningID := 0;
      ResetEvent(IdleEvent);
      Engine.Sequencing := True;
    end
    else
    begin
      RunningID := Requests.Get;
      if (RunningID <> 0) and Engine.LoadSequence(LoWord(RunningID)) then
      begin
        if RunningID <> IdleID then
          IdleLevel := Low(TIdleLevel);
        ResetEvent(IdleEvent);
        Engine.Sequencing := True;
      end
      else
        RunningID := 0;
    end;
  finally
    LeaveCriticalSection(CS);
  end;
end;

procedure TAssistantDirector.PlayIdle;
var
  Handled: Boolean;
  ThisLevel: Integer;
begin
  Handled := False;
  if Assigned(fOnIdle) then
  begin
    ThisLevel := Ord(IdleLevel) - Ord(Low(TIdleLevel)) + 1;
    try
      fOnIdle(Self, ThisLevel, Handled);
    finally
      fIdleLevel := TIdleLevel(Ord(Low(TIdleLevel)) + ThisLevel - 1);
    end;
  end;
  EnterCriticalSection(CS);
  try
    if not Handled then
      IdleID := Requests.Put(Actor.IndexOfAction(IdleLevel, 0))
    else
      IdleID := Requests[0];
    Dec(IdleRetryCount);
    if (IdleRetryCount = 0) and (IdleLevel < High(TIdleLevel)) then
      IdleLevel := Succ(IdleLevel)
    else
      AdjustIdleTimeout;
  finally
    LeaveCriticalSection(CS);
  end;
end;

procedure TAssistantDirector.SetEnabled(Value: Boolean);
begin
  if Enabled <> Value then
  begin
    EnterCriticalSection(CS);
    try
      fEnabled := Value;
      if not Enabled then
      begin
        ResetEvent(ActiveEvent);
        Requests.Enabled := False;
        StopAll;
      end
      else if not Actor.Empty then
      begin
        IdleLevel := Low(TIdleLevel);
        Requests.Enabled := True;
        SetEvent(ActiveEvent);
      end;
    finally
      LeaveCriticalSection(CS);
    end;
  end;
end;

procedure TAssistantDirector.SetAutoIdle(Value: Boolean);
begin
  if AutoIdle <> Value then
  begin
    fAutoIdle := Value;
    EnterCriticalSection(CS);
    try
      IdleLevel := Low(TIdleLevel);
      AdjustIdleTimeout;
      if Requests.Enabled and (Requests.Count = 0) then
      begin
        Requests.Enabled := False;
        Requests.Enabled := True;
      end;
    finally
      LeaveCriticalSection(CS);
    end;
  end;
end;

procedure TAssistantDirector.SetIdleLevel(Value: TIdleLevel);
var
  R: Integer;
begin
  if (IdleLevel <> Value) and (Value in [Low(TIdleLevel)..High(TIdleLevel)]) then
  begin
    EnterCriticalSection(CS);
    try
      fIdleLevel := Value;
      if (IdleLevel = Low(TIdleLevel)) and ((RunningID = 0) and (Requests.Count = 0)) then
        PlayAction(aaRestPose);
      Randomize;
      AdjustIdleTimeout;
    finally
      LeaveCriticalSection(CS);
    end;
  end;
  R := 5 * (Ord(High(TIdleLevel)) - Ord(IdleLevel) + 1);
  IdleRetryCount := R + Random(R);
end;

procedure TAssistantDirector.AdjustIdleTimeout;
begin
  if AutoIdle then
    IdleTimeout := MinIdlePeriod + Random(MaxIdlePeriod - MinIdlePeriod + 1)
  else
    IdleTimeout := INFINITE;
end;

procedure TAssistantDirector.ActorLoaded(Sender: TObject);
begin
  EnterCriticalSection(CS);
  try
    Engine.Initialize;
    if Enabled then
    begin
      Requests.Enabled := True;
      SetEvent(ActiveEvent);
    end;
    IdleLevel := Low(TIdleLevel);
  finally
    LeaveCriticalSection(CS);
  end;
end;

procedure TAssistantDirector.ActorUnloaded(Sender: TObject);
begin
  EnterCriticalSection(CS);
  try
    ResetEvent(ActiveEvent);
    Requests.Enabled := False;
    Engine.Finalize;
    Requests.Clear;
    RunningID := 0;
  finally
    LeaveCriticalSection(CS);
  end;
end;

procedure TAssistantDirector.SequenceStarted(Sender: TObject);
begin
  if (RunningID <> 0) and (RunningID <> IdleID) and Assigned(fOnPlayStart) then
    fOnPlayStart(Self, RunningID);
end;

procedure TAssistantDirector.SequenceStopped(Sender: TObject);
begin
  EnterCriticalSection(CS);
  try
    if RunningID = IdleID then
      IdleID := 0
    else if (RunningID <> 0) and Assigned(fOnPlayStop) then
      fOnPlayStop(Self, RunningID);
    SetEvent(IdleEvent);
  finally
    LeaveCriticalSection(CS);
  end;
end;

function TAssistantDirector.IsValidRequest(RequestID: TRequestID): Boolean;
begin
  EnterCriticalSection(CS);
  try
    Result := (RequestID <> 0) and (Requests.Query(RequestID) or
              ((RequestID = RunningID) and Engine.Sequencing));
  finally
    LeaveCriticalSection(CS);
  end;
end;

function TAssistantDirector.PlayIndex(Index: Word): TRequestID;
begin
  Result := Requests.Put(Index);
end;

function TAssistantDirector.PlayName(const Name: String): TRequestID;
begin
  Result := PlayIndex(Actor.IndexOfName(Name));
end;

function TAssistantDirector.PlayAction(Action: TActorAction): TRequestID;
begin
  Result := PlayIndex(Actor.IndexOfAction(Action, 0));
end;

function TAssistantDirector.PlayRandom(const ActionList: array of TActorAction): TRequestID;
type
  PActorActionArray = ^TActorActionArray;
  TActorActionArray = array[0..MaxShort] of TActorAction;
var
  List: PActorActionArray;
  Count, Index: Integer;
begin
  Result := 0;
  Count := High(ActionList) + 1;
  GetMem(List, Count * SizeOf(TActorAction));
  try
    for Index := 0 to Count - 1 do
      List[Index] := ActionList[Index];
    while (Count > 0) and (Result = 0) do
    begin
      Index := Random(Count);
      if List[Index] <> aaNone then
      begin
        Result := PlayAction(List[Index]);
        if Result = 0 then
        begin
          Dec(Count);
          for Index := Index downto Count - 1 do
            List[Index] := List[Index + 1];
        end;
      end
      else
        Break;
    end;
  finally
    FreeMem(List);
  end;
end;

procedure TAssistantDirector.Stop(RequestID: TRequestID);
begin
  if RunningID = RequestID then
  begin
    Engine.Sequencing := False;
    WaitForEventAndPaint(IdleEvent);
  end
  else
    Requests.Cancel(RequestID);
end;

procedure TAssistantDirector.StopAll;
begin
  Requests.Clear;
  if RunningID <> 0 then
  begin
    Engine.Sequencing := False;
    WaitForEventAndPaint(IdleEvent);
  end;
end;

function TAssistantDirector.WaitForRequest(RequestID: TRequestID): Boolean;
var
  CanWait: Boolean;
begin
  Result := True;
  while Result and IsValidRequest(RequestID) do
  begin
    EnterCriticalSection(CS);
    try
      CanWait := (RunningID = 0) or not Engine.SequenceIsInfinite;
    finally
      LeaveCriticalSection(CS);
    end;
    if CanWait then
      WaitForEventAndPaint(IdleEvent)
    else
      Result := False;
  end;
end;

function TAssistantDirector.Playing: Boolean;
begin
  EnterCriticalSection(CS);
  try
    Result := ((RunningID <> 0) and Engine.Sequencing) or (Requests.Count > 0);
  finally
    LeaveCriticalSection(CS);
  end;
end;

function TAssistantDirector.PlayingID: TRequestID;
begin
  if Engine.Sequencing then
    Result := RunningID
  else
    Result := 0;
end;

function TAssistantDirector.PlayIdling: Boolean;
begin
  Result := IsValidRequest(IdleID);
end;

function TAssistantDirector.IsPlaying(Actions: TActorActions): Boolean;
var
  Index: Integer;
  AnimInfo: PAnimationInfo;
begin
  Result := False;
  EnterCriticalSection(CS);
  try
    if (RunningID <> 0) and Engine.Sequencing then
    begin
      AnimInfo := Actor.AnimationInfo[LoWord(RunningID)];
      if (AnimInfo <> nil) and (AnimInfo^.Actions * Actions <> []) then
        Result := True;
    end;
    if not Result then
      for Index := 0 to Requests.Count - 1 do
      begin
        AnimInfo := Actor.AnimationInfo[LoWord(Requests[Index])];
        if (AnimInfo <> nil) and (AnimInfo^.Actions * Actions <> []) then
        begin
          Result := True;
          Break;
        end;
      end;
  finally
    LeaveCriticalSection(CS);
  end;
end;

function TAssistantDirector.ActionOfOrientation(Kind: TOrientedAction;
  const Origin, Target: TPoint): TActorAction;
type
  TDirX = (xNone, Left, Right);
  TDirY = (yNone, Down, Up);
const
  Actions: array[TOrientedAction, TDirX, TDirY] of TActorAction = (
    // oaGesture
    ((aaNone, aaGestureDown, aaGestureUp),
     (aaGestureLeft, aaGestureDownLeft, aaGestureUpLeft),
     (aaGestureRight, aaGestureDownRight, aaGestureUpRight)),
    // oaGlance
    ((aaNone, aaGlanceDown, aaGlanceUp),
     (aaGlanceLeft, aaGlanceDownLeft, aaGlanceUpLeft),
     (aaGlanceRight, aaGlanceDownRight, aaGlanceUpRight)),
    // oaLook
    ((aaNone, aaLookDown, aaLookUp),
     (aaLookLeft, aaLookDownLeft, aaLookUpLeft),
     (aaLookRight, aaLookDownRight, aaLookUpRight)),
    // oaMove
    ((aaNone, aaMoveDown, aaMoveUp),
     (aaMoveLeft, aaMoveDownLeft, aaMoveUpLeft),
     (aaMoveRight, aaMoveDownRight, aaMoveUpRight)));
var
  DirX: TDirX;
  DirY: TDirY;
begin
  if (Abs(Origin.x - Target.x) > (Actor.Width div 2)) or (Kind = oaMove) then
    if Origin.x > Target.x then
      DirX := Right
    else
      DirX := Left
  else
    DirX := xNone;
  if (Abs(Origin.y - Target.y) > (Actor.Height div 2)) or (Kind = oaMove) then
    if Origin.y > Target.y then
      DirY := Up
    else
      DirY := Down
  else
    DirY := yNone;
  Result := Actions[Kind, DirX, DirY];
  if Actor.IndexOfAction(Result, 0) = 0 then
  begin
    if Abs(Origin.x - Target.x) < Abs(Origin.y - Target.y) then
      DirX := xNone
    else
      DirY := yNone;
    Result := Actions[Kind, DirX, DirY];
  end;
end;

end.
